
/*

  The high-level plotting routines.

  Functions 'plot_...' are the entry points to this module.
  Functions 'draw_...', etc, are "private".

*/

#include <stdio.h>
#include "types.h"
#include "shapes.h"

char x_fmt [MAX_STR_SIZE], x_lab [MAX_STR_SIZE] ;
char y_fmt [MAX_STR_SIZE], y_lab [MAX_STR_SIZE] ;

LAYOUT layout ;

AXS_RANGE x_axis =
{
  0.0, 0.0,
  0.0, 0.0,
  0,
  x_fmt,
  x_lab,
  0
} ;

AXS_RANGE y_axis =
{
  0.0, 0.0,
  0.0, 0.0,
  0,
  y_fmt,
  y_lab,
  0
} ;

DISP d_x, d_y, n_x, n_y ;     /* unit vectors, for axes and their normals */

int num_pnts ;

/* Variables for plotting points... */

#define ABSENT 0

float x_coord [MAX_NUM_PNTS], y_coord [MAX_NUM_PNTS] ;
float x_error [MAX_NUM_PNTS], y_error [MAX_NUM_PNTS] ;

int point_type ;
int i_x_coord, i_x_error, i_y_coord, i_y_error ;

int absc_implicit ;
float absc_m, absc_c ;


void map (point, x_dat, y_dat)
POINT *point ;
float x_dat, y_dat ;

/*

  Map from a data point to a screen location

*/
{
  COORD x_min_scr = layout.bl_pos.x ;
  COORD y_min_scr = layout.bl_pos.y ;
  COORD x_max_scr = layout.tr_pos.x ;
  COORD y_max_scr = layout.tr_pos.y ;

  float x_min_dat = x_axis.min ;
  float y_min_dat = y_axis.min ;
  float x_max_dat = x_axis.max ;
  float y_max_dat = y_axis.max ;

  float x_int_dat = x_max_dat - x_min_dat ;
  float y_int_dat = y_max_dat - y_min_dat ;

  if (x_int_dat == 0.0)
  {
    fprintf (stderr, "ERROR: x-range is zero\n") ;
    exit (0) ;
  }
  else
    point->x = ((((x_min_scr * x_max_dat) - (x_max_scr * x_min_dat)) +
	         ((x_max_scr - x_min_scr) * x_dat)) / x_int_dat) + 0.5 ;

  if (y_int_dat == 0.0)
  {
    fprintf (stderr, "ERROR: y-range is zero\n") ;
    exit (0) ;
  }
  else
    point->y = ((((y_min_scr * y_max_dat) - (y_max_scr * y_min_dat)) +
	         ((y_max_scr - y_min_scr) * y_dat)) / y_int_dat) + 0.5 ;
}

void extract_gradations (r_start, r_offset, r_interval,
     		         num, maj_offset, maj_interval,
		         range)

float *r_start, *r_offset, *r_interval ;
int   *num, *maj_offset, *maj_interval ;
AXS_RANGE *range ;

/*

  Extract information about gradations from range structure.

  r_start    is the value of the ordinate (abscissa) at origin of axis.
  r_offset   is the offset from r_start of the first gradation (whether
	     it is major or minor).
  r_interval is the ordinate (absicissa) interval between any adjacent
             gradations (whether major or minor).
  num        is the total number of gradations to be printed.
  maj_offset is the index (starting from zero) of the first MAJOR
	     gradation, amongst the set of gradations to be printed.
  maj_interval 
	     is the interval at which major gradations occur.

*/
{
  float minimum  = range->min  ;
  float maximum  = range->max  ;
  float length   = maximum - minimum ;
  float eps      = 1.0E-4 * length ;
  int subdivs    = range->subd ;
  float anchor   = range->anch ;
  float min_intv = range->intv / subdivs ;
  float residue, offset ;
  int shift ;

  if (min_intv == 0)
  {
    fprintf (stderr, "ERROR: gradation interval is zero\n") ;
    exit (0) ;
  }

  offset   = anchor - minimum ;

  /* Put offset into range [-eps, min_intv] */

  if (offset < (-eps))
    shift = 1 + ((-offset) / min_intv) ;
  else
    shift = - (offset / min_intv) ;
  offset += shift * min_intv ;

  residue = length - offset ;

  *r_start    = minimum ;
  *r_offset   = offset ;
  *r_interval = min_intv ;


  *num        = (residue > (-eps)) ? (1 + (residue / min_intv)) : 0 ;
  *maj_offset = (shift <= 0) ? ((-shift) % subdivs) :
		               ((subdivs - (shift % subdivs)) % subdivs) ;
  *maj_interval = subdivs ;
}


void draw_title (title) 
char *title ;
{ 
  put_string (title, &layout.title_pos, layout.title_siz) ;
}

void draw_major (point, norm, do_number, text, number, layout)
POINT *point ;
DISP  *norm  ;
int   do_number ;
char  *text ;
float number ;
AXS_LAYOUT *layout ;
/*
  
  Draw a major gradation on an axis.

  point     is the point at which the grad meets the axis.
  norm      is a unit vector, in which direction the line should be drawn.
  do_number is set if we are to print an associated value.
  text, number
	    are the format text and the value to be printed.
  layout    is the axis layout structure.

*/
{
  WIDTH len = layout->maj_wid ;
  POINT end, num_pos ;
  char string [MAX_STR_SIZE] ;

  end.x = point->x + (len * norm->x) ;
  end.y = point->y + (len * norm->y) ;
  put_line (point, &end) ;

  if (do_number)
  {
    num_pos.x = point->x + layout->num_dsp.x ;
    num_pos.y = point->y + layout->num_dsp.y ;
    sprintf (string, text, number) ;
    put_string (string, &num_pos, layout->num_siz) ;
  }
}

void draw_minor (point, norm, layout)
POINT *point ;
DISP  *norm  ;
AXS_LAYOUT *layout ;
/*
  
  Draw a minor gradation on an axis.

  point     is the point at which the grad meets the axis.
  norm      is a unit vector, in which direction the line should be drawn.
  layout    is the axis layout structure.

*/
{
  WIDTH len = layout->min_wid ;
  POINT end ;

  end.x = point->x + (len * norm->x) ;
  end.y = point->y + (len * norm->y) ;
  put_line (point, &end) ;
}

void draw_label (end, label, layout)
POINT *end ;
char  *label ;
AXS_LAYOUT *layout ;
/*

  Draw label on an axis

  end       is the point at the end of the axis, for orientation.
  label     is the text.
  layout    is the axis layout structure.

*/
{
  POINT point ;

  point.x = end->x + layout->lab_dsp.x ;
  point.y = end->y + layout->lab_dsp.y ;

  put_string (label, &point, layout->lab_siz) ;
}


void draw_axis (at, len, dirn, norm, layout, range, do_number, do_label)
POINT *at ;
COORD len ;
DISP *dirn, *norm ;
AXS_LAYOUT *layout ;
AXS_RANGE  *range ;
int do_number, do_label ;
/*

  'at'   is point of origin of axis on screen.
  'len'  is its length on screen
  'dirn' is a unit vector, giving it's direction on screen.
  'norm' is a unit vector, giving the gradation direction.
  'layout', 'range'       are the data structures associted with the axis.
  'do_number', 'do_label' are set if axis is to be numbered, labelled.

  This routine ought to be able to call mapping routines from
  data coordinate to screen coordinate which were common with those used
  by the data point plotting routines.  Unfortunately I arranged
  my data structures badly, and this can't be done easily, without
  a significant rewrite.

*/
{
  POINT end ;

  end.x = at->x + (len * dirn->x) ;
  end.y = at->y + (len * dirn->y) ;

  /* Draw basic axis */

  put_line (at, &end) ;

  /* Draw gradations, and number them */
  {
    float r_start, r_offset, r_interval ;
    int   num, maj_offset, maj_interval ;
    int   scale = len / (range->max - range->min) ;
    int i ;

    extract_gradations (&r_start, &r_offset, &r_interval,
		        &num, &maj_offset, &maj_interval,
		        range) ;
    for (i = 0; i < num; i++)
    {
      float r_disp = r_offset + (i * r_interval) ;
      float r_abs  = r_start  +  r_disp ;
      COORD offset = (scale * r_disp) + 0.5 ;
      POINT point ;

      point.x = at->x + (offset * dirn->x) ;
      point.y = at->y + (offset * dirn->y) ;

      if ((i % maj_interval) == maj_offset)
	draw_major (&point, norm, do_number, range->fmt, r_abs, layout) ;
      else
	draw_minor (&point, norm, layout) ;
    }
  }

  /* Label it */

  if (do_label)
    draw_label (&end, range->lab, layout) ;
}

void draw_edge_pair (at_1, at_2, len, dirn, norm, layout, range)
POINT *at_1, *at_2 ;
COORD len ;
DISP *dirn, *norm ;
AXS_LAYOUT *layout ;
AXS_RANGE  *range ;
{
  DISP neg_norm ;

  neg_norm.x = -norm->x ;
  neg_norm.y = -norm->y ;

  draw_axis (at_1, len, dirn, norm, layout, range, 1, 1) ;
  if (range->dup)
    draw_axis (at_2, len, dirn, &neg_norm, layout, range, 0, 0) ;
}

void draw_centre (x, y, type)
float x, y ;
int type ;
{
  POINT centre ;
  POINT poly [MAX_NUM_VRTS] ;
  int shape_index = type > 0 ? type - 1 : - type - 1 ;
  int colour      = type > 0 ? 1 : 21 ;
  int num_vertices = num_vrts [shape_index] ;
  DISP *shape      = shapes   [shape_index] ;
  COORD r = layout.pnt_wid / 2 ;
  int i ;

  map (&centre, x, y) ;

  for (i = 0; i < num_vertices; i++)
  {
    poly [i].x = centre.x + (shape [i].x * r) ;
    poly [i].y = centre.y + (shape [i].y * r) ;
  }
  put_poly (poly, num_vertices, colour) ;
}

void draw_error (x, y, err, dirn, norm) 
float x, y, err ;
DISP *dirn, *norm ;
{
  float x_t, y_t, x_b, y_b ;
  POINT top, bot, top_l, top_r, bot_l, bot_r ;
  COORD r = layout.pnt_wid / 2 ;

  x_t = x + (dirn->x * err) ;
  y_t = y + (dirn->y * err) ;
  x_b = x - (dirn->x * err) ;
  y_b = y - (dirn->y * err) ;

  map (&top, x_t, y_t) ;
  map (&bot, x_b, y_b) ;

  top_l.x = top.x - (norm->x * r) ;
  top_l.y = top.y - (norm->y * r) ;
  top_r.x = top.x + (norm->x * r) ;
  top_r.y = top.y + (norm->y * r) ;
  bot_l.x = bot.x - (norm->x * r) ;
  bot_l.y = bot.y - (norm->y * r) ;
  bot_r.x = bot.x + (norm->x * r) ;
  bot_r.y = bot.y + (norm->y * r) ;

  put_line (&bot, &top) ;
  put_line (&bot_l, &bot_r) ;
  put_line (&top_l, &top_r) ;
}


void draw_point (x, y, do_err_x, err_x, do_err_y, err_y, type)
float x, y, err_x, err_y ;
int do_err_x, do_err_y, type ;
{
  if (type != 0)
  {
    if (do_err_x)
      draw_error (x, y, err_x, &d_x, &n_x) ;
    if (do_err_y)
      draw_error (x, y, err_y, &d_y, &n_y) ;

    draw_centre (x, y, type) ;
  }
}

void plot_box (title)
char *title ;
{
  POINT end_x, end_y ;          /* end of x, y axis */

  COORD len_x = layout.tr_pos.x - layout.bl_pos.x ;
  COORD len_y = layout.tr_pos.y - layout.bl_pos.y ;
  
  end_x.x = layout.tr_pos.x ;
  end_x.y = layout.bl_pos.y ;

  end_y.x = layout.bl_pos.x ;
  end_y.y = layout.tr_pos.y ;

  d_x.x = 1 ;
  d_x.y = 0 ;
  d_y.x = 0 ;
  d_y.y = 1 ;
  n_x.x = 0 ;
  n_x.y = len_y > 0 ? 1 : -1 ;
  n_y.x = len_x > 0 ? 1 : -1 ;
  n_y.y = 0 ;

  put_begin () ;

  draw_edge_pair (&(layout.bl_pos), &end_y, len_x, &d_x, &n_x,
		  &(layout.x_lay), &x_axis) ;
  draw_edge_pair (&(layout.bl_pos), &end_x, len_y, &d_y, &n_y,
		  &(layout.y_lay), &y_axis) ;

  draw_title (title) ;
}

void plot_begin (data_header)
DAT_HDR *data_header ;

/*

  Initialise point counter, and some flags and indicies.

*/
{

  num_pnts = 0 ;

  point_type = data_header->pnt_fmt ;

  /* Indicies of various things in an input 'point' record */

  {
    int i = 0 ;

    if (data_header->imp)
    {
      absc_implicit = 1 ;
      absc_m       = data_header->form.m ;
      absc_c       = data_header->form.c ;
      i_x_coord    = ABSENT ;
    }
    else
    {
      i_x_coord = i ;
      i++ ;
    }

    if (data_header->err_x)
    {
      i_x_error = i ;
      i++ ;
    }
    else
      i_x_error = ABSENT ;

    i_y_coord = i ;
    i++ ;

    if (data_header->err_y)
    {
      i_y_error = i ;
      i++ ;
    }
    else
      i_y_error = ABSENT ;
  }
}

void plot_end (data_header)
DAT_HDR *data_header ;

/*

  Plot the data.

*/
{
  int i ;

  /* Do spline interpolation, if required, first */

  if (data_header->spl)
  {
    fprintf (stderr, "ERROR: Spline interpolation not implemented.\n") ;
    exit (0) ;
  }

  for (i = 0; i < num_pnts; i++)
    draw_point (x_coord [i], y_coord [i],
		data_header->err_x, x_error [i],
		data_header->err_y, y_error [i],
		data_header->pnt_fmt) ;
}


void plot_point (point)
float point [] ;
{
  if (num_pnts < MAX_NUM_PNTS)
  {
    if (absc_implicit)
      x_coord [num_pnts] = (absc_m * num_pnts) + absc_c ;
    else
      x_coord [num_pnts] = point [i_x_coord] ;
    x_error [num_pnts] = point [i_x_error] ;
    y_coord [num_pnts] = point [i_y_coord] ;
    y_error [num_pnts] = point [i_y_error] ;
    fprintf (stderr, "x, y, err = %f %f %f\n", x_coord [num_pnts],
	     y_coord [num_pnts], y_error [num_pnts]) ;
    num_pnts++ ;
  }
  else
  {
    fprintf (stderr, "ERROR: Too many data points to plot\n") ;
    exit (0) ;
  }
}
